/*! ********************************************
 * GBEventDispatcher.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.event 
{
import flash.display.Sprite;
import flash.events.Event;
import flash.events.EventDispatcher;
import flash.utils.ByteArray;
import flash.utils.getTimer;
import sora.graybox.GrayBox;
import sora.graybox.IGrayBoxModule;
import sora.graybox.MidiEvent;
import sora.graybox.ProcessData;

public final class GBEventDispatcher extends EventDispatcher implements IGrayBoxModule 
{
	private var _sprite     :Sprite = null;
	//
	public var midiInPorts  :int = 0xFFFF;
	private var _events     :GBEvent = null;
	private var _frameRate  :Number = 30.0;
	private var _sampleRate :Number = 44100.0;
	
	private var _smplK    :Number = _sampleRate / 1000.0;
	private var _srpfr    :int = int(_sampleRate / _frameRate);
	private var _procTime :int = 0;
	
	public var dispSysExEvent:Boolean = false;
	public var dispMetaMessage:Boolean = false;
	
	public function setFrameRate(rate:Number):void 
	{
		_frameRate = rate;
		_srpfr = int(_sampleRate / _frameRate);
	}
// Eventhandler
	private function _enterFrame(event:Event):void 
	{
		var _procDelay:int = 0;
	// Calc DelayTime -->
		if (_procTime > 0)
		{
			_procDelay = getTimer() - _procTime;
			_procDelay = int(_procDelay * _smplK);
			_procTime = 0;
		}
	// dipatch Events -->
		var _prev:GBEvent = _events,
			_crnt:GBEvent = _prev.next;
		while (_crnt != null)
		{
			if (_crnt.fr <= 0)
			{
				_prev.next = _crnt.next;
				_crnt.next = null;
				dispatchEvent(_crnt);
				_crnt = _prev.next;
				continue;
			}
			_crnt.fr -= _srpfr + _procDelay;
			_prev = _crnt;
			_crnt = _crnt.next;
		}
	}
// Constructor
	public function GBEventDispatcher():void 
	{
		super();
		_events = new GBEvent(GBEvent.MIDI_CHANNEL_MESSAGE);
		_sprite = new Sprite();
	}
// Interface - IfGrayBoxModule -----
	public function updateSampleRate(sampleRate:Number):void 
	{
		_sampleRate = sampleRate;
		_smplK = _sampleRate / 1000.0;
		_srpfr = int(_sampleRate / _frameRate);
	}
	public function updateAudioChannels(ch:int):void{}
	public function updateProcessBlockSize(size:int):void{}
	public function resetModule():void { _events.next = null; }
	public function initiModule():void { _events.next = null; }
	public function connectModule(host:GrayBox):void
	{
		_events.next = null;
		_frameRate = (_sprite.stage != null)?_sprite.stage.frameRate : 60.0;
		_sampleRate = host.getSampleRate();
		_srpfr = int(_sampleRate / _frameRate);
		_procTime = 0;
		_sprite.removeEventListener(Event.ENTER_FRAME, _enterFrame, false);
		_sprite.addEventListener(Event.ENTER_FRAME, _enterFrame, false, 0, true);
	}
	public function disconnectModule():void
	{
		_events.next = null;
		_procTime = 0;
		_sprite.removeEventListener(Event.ENTER_FRAME, _enterFrame, false);
	}
	public function updateBypassStatus(bypass:Boolean):void { _events.next = null; }
	public function startProcess():void
	{
		_events.next = null;
		_procTime = 0;
		_sprite.removeEventListener(Event.ENTER_FRAME, _enterFrame, false);
		_sprite.addEventListener(Event.ENTER_FRAME, _enterFrame, false, 0, true);
	}
	public function stopProcess():void
	{
		_events.next = null;
		_procTime = 0;
		_sprite.removeEventListener(Event.ENTER_FRAME, _enterFrame, false);
	}
	public function suspendProcess():void { _events.next = null; }
	public function resumeProcess():void { _events.next = null; }
	public function process(procData:ProcessData):void 
	{
		var _list:MidiEvent = procData.midiEventList.next;
		var _ev:GBEvent = _events;
		while (_ev.next != null)_ev = _ev.next; // LastEvent
		var _gbEventList:GBEvent = _ev;
		var _frameDelay:int = int(procData.audioLatency * _smplK);
		// process
		while (_list != null)
		{
			if ((_list.midiPorts & midiInPorts) == 0){ _list = _list.next; continue; }
			// meta message (SMF)
			if (_list.type >= 0xFF00)
			{
				if (!dispMetaMessage || _list.type > 0xFF7F){ _list = _list.next; continue; }
				_ev = new GBEvent(GBEvent.SMF_META_MESSAGE);
				_ev.fr = _list.offsetFrames + _frameDelay;
				_ev.messageType = _list.type;
				_ev.value0 = _list.value0;
				if (_list.data != null)
				{
					_ev.data = new ByteArray();
					_ev.data.writeBytes(_list.data);
					_ev.data.position = 0;
				}
				_gbEventList.next = _ev;
				_gbEventList = _ev;
			}
			// system message
			else if (_list.type >= 0xF0)
			{
				if (!dispSysExEvent){ _list = _list.next; continue; }
				_ev = new GBEvent(GBEvent.MIDI_SYSTEM_MESSAGE);
				_ev.fr = _list.offsetFrames + _frameDelay;
				_ev.messageType = _list.type;
				_ev.channel = _list.channel;
				_ev.value0 = _list.value0;
				_ev.value1 = _list.value1;
				if (_list.data != null)
				{
					trace(_list.data.length);
					_ev.data = new ByteArray();
					_ev.data.writeBytes(_list.data);
					_ev.data.position = 0;
				}
				_gbEventList.next = _ev;
				_gbEventList = _ev;
			}
			// channel message
			else if (_list.type >= 0x80)
			{
				_ev = new GBEvent(GBEvent.MIDI_CHANNEL_MESSAGE);
				_ev.fr = _list.offsetFrames + _frameDelay;
				_ev.messageType = _list.type;
				_ev.channel = _list.channel;
				_ev.value0 = _list.value0;
				_ev.value1 = _list.value1;
				_gbEventList.next = _ev;
				_gbEventList = _ev;
			}
			// next
			_list = _list.next;
		}
		_procTime = getTimer();
	}
}
}